from .version import __version__

from .layer import CoralOrdinal
from .loss import OrdinalCrossEntropy
from .metrics import MeanAbsoluteErrorLabels
from .activations import ordinal_softmax

# if somebody does "from somepackage import *", this is what they will
# be able to access:
__all__ = [
    "CoralOrdinal",
    "MeanAbsoluteErrorLabels",
    "OrdinalCrossEntropy",
    "ordinal_softmax",
]



import tensorflow as tf
from tensorflow.python.keras import activations


@tf.keras.utils.register_keras_serializable(package="coral_ordinal")
class CoralOrdinal(tf.keras.layers.Layer):

    # We skip input_dim/input_shape here and put in the build() method as recommended in the tutorial,
    # in case the user doesn't know the input dimensions when defining the model.
    def __init__(self, num_classes: int, activation=None, **kwargs):
        """Ordinal output layer, which produces ordinal logits by default.

        Args:
          num_classes: how many ranks (aka labels or values) are in the ordinal variable.
          activation: (Optional) Activation function to use. The default of None produces
            ordinal logits, but passing "ordinal_softmax" will cause the layer to output
            a probability prediction for each label.
        """

        # Via Dense Layer code:
        # https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/layers/core.py#L1128
        if "input_shape" not in kwargs and "input_dim" in kwargs:
            kwargs["input_shape"] = (kwargs.pop("input_dim"),)

        # Pass any additional keyword arguments to Layer() (i.e. name, dtype)
        super(CoralOrdinal, self).__init__(**kwargs)
        self.num_classes = num_classes
        self.activation = activations.get(activation)

    # Following https://www.tensorflow.org/guide/keras/custom_layers_and_models#best_practice_deferring_weight_creation_until_the_shape_of_the_inputs_is_known
    def build(self, input_shape):

        # Single fully-connected neuron - this is the latent variable.
        num_units = 1

        # I believe glorot_uniform (aka Xavier uniform) is pytorch's default initializer, per
        # https://pytorch.org/docs/master/generated/torch.nn.Linear.html
        # and https://www.tensorflow.org/api_docs/python/tf/keras/initializers/GlorotUniform
        self.fc = self.add_weight(
            shape=(input_shape[-1], num_units),
            # Need a unique name if there are multiple coral_ordinal layers.
            name=self.name + "_latent",
            initializer="glorot_uniform",
            # Not sure if this is necessary:
            dtype=tf.float32,
            trainable=True,
        )

        # num_classes - 1 bias terms, defaulting to 0.
        self.linear_1_bias = self.add_weight(
            shape=(self.num_classes - 1,),
            # Need a unique name if there are multiple coral_ordinal layers.
            name=self.name + "_bias",
            initializer="zeros",
            # Not sure if this is necessary:
            dtype=tf.float32,
            trainable=True,
        )

    # This defines the forward pass.
    def call(self, inputs):
        fc_inputs = tf.matmul(inputs, self.fc)

        logits = fc_inputs + self.linear_1_bias

        if self.activation is None:
            outputs = logits
        else:
            # Not yet tested:
            outputs = self.activation(logits)

        return outputs

    # This allows for serialization supposedly.
    # https://www.tensorflow.org/guide/keras/custom_layers_and_models#you_can_optionally_enable_serialization_on_your_layers
    def get_config(self):
        config = super(CoralOrdinal, self).get_config()
        config.update({"num_classes": self.num_classes})
        return config



import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package="coral_ordinal")
def ordinal_softmax(x, axis=-1):
    """Convert the ordinal logit output of CoralOrdinal() to label probabilities.

    Args:
      x: Logit output of the CoralOrdinal() layer.
      axis: Not yet supported.
    """

    # Number of columns is the number of classes - 1
    num_classes = x.shape[1] + 1

    # Convert the ordinal logits into cumulative probabilities.
    cum_probs = tf.math.sigmoid(x)

    # Create a list of tensors.
    probs = []

    # First, get probability predictions out of the cumulative logits.
    # Column 0 is Probability that y > 0, so Pr(y = 0) = 1 - Pr(y > 0)
    # Pr(Y = 0) = 1 - s(logit for column 0)
    probs.append(1.0 - cum_probs[:, 0])

    # For the other columns, the probability is:
    # Pr(y = k) = Pr(y > k) - Pr(y > k - 1)
    if num_classes > 2:
        for val in range(1, num_classes - 1):
            probs.append(cum_probs[:, val - 1] - cum_probs[:, val])

    # Special handling of the maximum label value.
    probs.append(cum_probs[:, num_classes - 2])

    # Combine as columns into a new tensor.
    probs_tensor = tf.concat(tf.transpose(probs), axis=1)

    return probs_tensor



import tensorflow as tf
from tensorflow.keras import backend as K


@tf.keras.utils.register_keras_serializable(package="coral_ordinal")
class MeanAbsoluteErrorLabels(tf.keras.metrics.Metric):
    """Computes mean absolute error for ordinal labels."""

    def __init__(self, name="mean_absolute_error_labels", **kwargs):
        """Creates a `MeanAbsoluteErrorLabels` instance."""
        super(MeanAbsoluteErrorLabels, self).__init__(name=name, **kwargs)
        self.maes = self.add_weight(name="maes", initializer="zeros")
        self.count = self.add_weight(name="count", initializer="zeros")

    def update_state(self, y_true, y_pred, sample_weight=None):
        """Computes mean absolute error for ordinal labels.

        Args:
          y_true: Labels (int).
          y_pred: Cumulative logits from CoralOrdinal layer.
          sample_weight (optional): Not implemented.
        """

        if sample_weight:
            raise NotImplementedError

        # Predict the label as in Cao et al. - using cumulative probabilities
        # cum_probs = tf.map_fn(tf.math.sigmoid, y_pred)

        # Calculate the labels using the style of Cao et al.
        # above_thresh = tf.map_fn(lambda x: tf.cast(x > 0.5, tf.float32), cum_probs)

        # Skip sigmoid and just operate on logit scale, since logit > 0 is
        # equivalent to prob > 0.5.
        above_thresh = tf.cast(y_pred > 0.0, tf.float32)

        # Sum across columns to estimate how many cumulative thresholds are passed.
        labels_v2 = tf.reduce_sum(above_thresh, axis=1)

        y_true = tf.cast(y_true, y_pred.dtype)

        # remove all dimensions of size 1 (e.g., from [[1], [2]], to [1, 2])
        y_true = tf.squeeze(y_true)

        self.maes.assign_add(tf.reduce_mean(tf.abs(y_true - labels_v2)))
        self.count.assign_add(tf.constant(1.0))

    def result(self):
        return tf.math.divide_no_nan(self.maes, self.count)

    def reset_state(self):
        """Resets all of the metric state variables at the start of each epoch."""
        K.batch_set_value([(v, 0) for v in self.variables])

    def get_config(self):
        """Returns the serializable config of the metric."""
        config = {}
        base_config = super().get_config()
        return {**base_config, **config}


"""
# WIP
def MeanAbsoluteErrorLabels_v2(y_true, y_pred):
  # There will be num_classes - 1 cumulative logits as columns of the tensor.
  num_classes = y_pred.shape[1] + 1
  
  probs = logits_to_probs(y_pred, num_classes)

# RootMeanSquaredErrorLabels
"""



from typing import Optional

import functools
import tensorflow as tf
import numpy as np


def _label_to_levels(labels: tf.Tensor, num_classes: int) -> tf.Tensor:
    # Original code that we are trying to replicate:
    # levels = [1] * label + [0] * (self.num_classes - 1 - label)
    # This function uses tf.sequence_mask(), which is vectorized. Avoids map_fn()
    # call.
    return tf.sequence_mask(labels, maxlen=num_classes - 1, dtype=tf.float32)


def _ordinal_loss_no_reduction(
    logits: tf.Tensor, levels: tf.Tensor, importance: tf.Tensor
) -> tf.Tensor:
    """Compute ordinal loss without reduction."""
    losses = -tf.reduce_sum(
        (
            tf.math.log_sigmoid(logits) * levels
            + (tf.math.log_sigmoid(logits) - logits) * (1.0 - levels)
        )
        * importance,
        axis=1,
    )
    return losses


# The outer function is a constructor to create a loss function using a certain number of classes.
@tf.keras.utils.register_keras_serializable(package="coral_ordinal")
class OrdinalCrossEntropy(tf.keras.losses.Loss):
    def __init__(
        self,
        num_classes: Optional[int] = None,
        importance_weights=None,
        from_type: str = "ordinal_logits",
        name: str = "ordinal_crossentropy",
        **kwargs,
    ):
        """Cross-entropy loss designed for ordinal outcomes.

        Args:
          num_classes: number of ranks (aka labels or values) in the ordinal variable.
            This is optional; can be inferred from size of y_pred at runtime.
          importance_weights: (Optional) importance weights for each binary classification task.
          from_type: one of "ordinal_logits" (default), "logits", or "probs".
            Ordinal logits are the output of a CoralOrdinal() layer with no activation.
            (Not yet implemented) Logits are the output of a dense layer with no activation.
            (Not yet implemented) Probs are the probability outputs of a softmax or ordinal_softmax layer.
          name: name of layer
          **kwargs: keyword arguments passed to Loss().
        """
        super().__init__(name=name, **kwargs)

        self.num_classes = num_classes
        self.importance_weights = importance_weights
        self.from_type = from_type

    # Following https://www.tensorflow.org/api_docs/python/tf/keras/losses/Loss
    def call(self, y_true: tf.Tensor, y_pred: tf.Tensor):

        # Ensure that y_true is the same type as y_pred (presumably a float).
        y_pred = tf.convert_to_tensor(y_pred)
        y_true = tf.cast(y_true, y_pred.dtype)
        if self.num_classes is None:
            self.num_classes = int(y_pred.get_shape().as_list()[1]) + 1

        # Convert each true label to a vector of ordinal level indicators.
        tf_levels = _label_to_levels(tf.squeeze(y_true), self.num_classes)

        if self.importance_weights is None:
            importance_weights = tf.ones(self.num_classes - 1, dtype=tf.float32)
        else:
            importance_weights = tf.cast(self.importance_weights, dtype=tf.float32)

        if self.from_type == "ordinal_logits":
            loss = _ordinal_loss_no_reduction(y_pred, tf_levels, importance_weights)
        elif self.from_type == "probs":
            raise NotImplementedError("not yet implemented")
        elif self.from_type == "logits":
            raise NotImplementedError("not yet implemented")
        else:
            raise Exception(
                "Unknown from_type value "
                + self.from_type
                + " in OrdinalCrossEntropy()"
            )
        if self.reduction == tf.keras.losses.Reduction.NONE:
            return loss
        elif self.reduction in [
            tf.keras.losses.Reduction.AUTO,
            tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE,
        ]:
            return tf.reduce_mean(loss)
        elif self.reduction == tf.keras.losses.Reduction.SUM:
            return tf.reduce_sum(loss)
        else:
            raise Exception(f"{self.reduction} is not a valid reduction.")

    def get_config(self):
        config = {
            "num_classes": self.num_classes,
            "importance_weights": self.importance_weights,
            "from_type": self.from_type,
        }
        base_config = super().get_config()
        return {**base_config, **config}



import setuptools
from setuptools import setup
import os
import sys

_here = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(os.path.join(_here, 'coral_ordinal', 'version.py')) as f:
    exec(f.read(), version)
    
    
with open(os.path.join(_here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = 'coral-ordinal',
    url = 'https://github.com/ck37/coral-ordinal',
    author = 'Chris Kennedy, Stephen Matthews, Georg M. Goerg',
    author_email = 'chrisken@gmail.com',
    packages = setuptools.find_packages(),
    install_requires = ['numpy', 'tensorflow>=2.2'],
    version = version['__version__'],
    long_description_content_type = "text/markdown",
    license = 'MIT',
    description = 'Tensorflow Keras implementation of CORAL ordinal regression output layer, loss, activation, and metrics',
    long_description = long_description,
    python_requires = '>=3.6',
)
